#! /usr/local/bin/perl

# Copyright (c) 1996-2000 The Regents of the University of California.
# All Rights Reserved.
# 
# Permission to use, copy, modify, and distribute this software and its
# documentation for educational, research and non-profit purposes,
# without fee, and without a written agreement is hereby granted,
# provided that the above copyright notice, this paragraph and the
# following three paragraphs appear in all copies.
# 
# Permission to incorporate this software into commercial products may
# be obtained from the Office of Technology Licensing, 2150 Shattuck
# Avenue, Suite 510, Berkeley, CA  94704.
# 
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
# ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
# CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
# UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

# $Id: qdata.pl,v 1.40 2010/01/08 21:57:27 doug Exp $ 

# qdata - extract MiniSEED / Telemetry SEED) data from various archives.
#
# Data can be located in any of the following places:
#   1.	Current Telemetry directories.
#   2.	Dialup directories (fill in data holes in telemetry data).
#   3.	Extracted daily directories, located on:
# 	a.  Active disk (migrated back from archive).
# 	b.  Stage disk (not migrated to archive yet).
#	c.  New archive.
# 	c.  Old archive.
#
#   1.  Archive dir:
# 	Assumed to contain directory years which contains daily directories.
#	Files in daily directories have filenames in archive format.
#
#   2.  Stage dir:
# 	Assumed to contain daily directories.
#	Files in daily directories have filenames in archive format.
#
#   3.  Active dir:
# 	Assumed to contain daily directories that superceed those in archive.
#	Files in daily directories have filenames in archive format.
#	This directory is available to the data administrator to explicitly 
#	place a copy of one or more archive daily directories on faster media.
#
#   4.  Telemetry directories:
#	Assumed to contain station directories with a DATALOG structure 
# 	consisting of:
# 		STATION.NET/CHANNEL[.LOC].D/file
# 	where:
#		STATION,CHANNEL are uppercase station and channel names.
#
#   5.  Dialup directories:
#	Assumed to contain demultiplexed station.network.channel[.location] 
#	files in dialup file format.
#
# Year directories have a filename format of:
# 	year
#   where:
# 	year = 4 digit year (eg 1995).
#
# Data files will be located in subdirectories (except for yearly archives)
# which are divided by doy or month,
# Directory filename formats:
# 	year.doy
# 	year.month
#   where:
# 	year = 4 digit year (eg 1995).
# 	doy = 3 digit day-of-year (eg 023).
#	month = 2 digit month-of-year (eg 05).
#
# Files in archive have a filename format of:
# 	station.net.channel[.loc].D.year.doy
# 	station.net.channel[.loc].D.year.month
# 	station.net.channel[.loc].D.year
#   where:
#	station,net,channel,loc are UPPERCASE station, network, 
#		channel, and location names.
# 	year.doy are defined above.
# 	year.month are defined above.
# 	year are defined above.
#
# Files in dialup format have a filename format of:
# 	station.net.channel[.loc].D.year.doy.hhmm
#   where:
#	station,net,channel are UPPERCASE station, network, and channel names.
# 	year.doy are defined above.
# 	hhmm are the hour and minute of the start of data (eg 0209).
# 
# A separate output file is created for each station/channel.
# Output files are named:
#	station.net.channel.[.loc]D.year.doy.hhmm
#	station.channel.D.year.doy.hhmm		(old format)
#    where:
#	year.doy.hhmm is the timestamp of the beginning request interval.
#	Note that this is NOT necessarily the timestamp of the first data point.

require 5.002;
use strict;
use Getopt::Std;
use Qtime;

use vars qw ( $VERSION $DFLT_NET $SDELIM $CDELIM $NODATA );
use vars qw ( $cmdname $newstyle $oldstyle $stdfmt );
use vars qw ( %netinfo @wc_list @dflt_stations $date_fmt %reject_sncls );
use vars qw ( $chanfile $chanpath $netsfile $netspath $libdir $qmerge $caldate );
use vars qw ( $opt_O $opt_D $opt_C $opt_P $opt_R $opt_T $opt_I $opt_A );
use vars qw ( $opt_h $opt_d $opt_f $opt_p $opt_f $opt_t $opt_s $opt_n $opt_o $opt_l $opt_a );
use vars qw ( $debug $current );

$VERSION = "1.40 (2010.008)";

########################################################################
#	print_syntax	- print syntax and exit.
########################################################################
sub print_syntax {
    my($cmdname) = @_;
    my($all_stations_string,$backup_stations_string);

    printf	
"$cmdname version $VERSION
$cmdname - extract an interval of Quanterra data from its various archives.
Syntax:
$cmdname   -f start_time  -t end_time | -s span_interval [-o output_file] [-n]
	[-a] [-A] [-T | -I] [-C] [-B] [-E] [-O] [-P] [-p] [-R reject_sncl_list] 
	[-D j|m] [-l libdir] [-d debug_value] [-h] 
	[sncl_list]
where:
	-f start_time
		Specifies start time for data
	-t end_time
		Specifies end time for data
	-s span_interval
		Span interval.  Alternate way of specifying end time.
		Interval can be an integer followed immediately by
		S, M, H, or d for seconds, minutes, hours, or days.
	-o outfile
		Specifies the output file to contain all of the data from
		this request.  If the outfile is '-', the data is written
		to stdout.  If this option is not used, data for each 
		station-channel is written to a separate file whose name
		is generated from the station, network, channel, and start
		time of the data request.
	-n	No data output.  Run extraction for messages only.
	-a	Use all data locations (archive and prelim).
	-A	Output data in ASCII instead of binary MiniSEED.
	-T | -I	Trim data to the specified timeperiod (-T), or include
		all record containing the specified time interval (-I).
		Default is -T.
	-C	Get current telemetry data ONLY.
	-B	Get current telemetry data from backup directories.
	-E	OBSOLETE
		Get EM (electro-magnetic) high-frequency (40 Hz) data.
	-O	Generate old-style filenames without net code.
	-P	Prints first date for preliminary data (data that has not been
		reviewed for timing problems).
	-p	Prints extraction cmd instead of executing it.
	-R reject_sncl_list
		Specifies a comma-delimited list of sncls to reject.
		The sncls may contain wildcards if quoted.
	-D j|m	Prints dates in julian(day-of-year) or month/day format
		Default format is julian.
	-l libdir
		Specifies alternate directory containing the configuration 
		files qdata.nets and qdata.channels.
	-d n	Debug value n.  See qmerge for details.
	-h	Help - prints this help message.
	sncl_list
		A list of one or more scnl tokens, where each token represents
			station.network.channel.location
		Each component may be a single identifier (eg BKS) or a 
		comma-delimited list. (eg BKS,CMB)
		The station, channel, or location may be wildcarded.
		The location token may be omitted.

Notes:
1.  A complete timespan (using -f and Either -t or -s) MUST be specified.
2.  SEED channel names:	Common names:
	----------------------------------------
	HHZ, HHN, HHE	VSP channels
	BHZ, BHN, BHE	VBB channels
	LHZ, LHN, LHE	LP channels
	VHZ, VHN, VHE	VLP channels
	UHZ, UHN, UHE	ULP channels
	HLZ, HLN, HLE	Strong Motion channels
	
Examples:
	$cmdname -f 1998.176,11:43 -s 10M MHC.BK.BHZ,BHN,BHE
	$cmdname -f 1998.176,11:43 -s 10M 'MHC.BK.BH*'
		Get 10 minutes of VBB data from station MHC.BK.
	$cmdname -f 1998.176,11:43 -s 10M 'MHC,SAO.BK.BH*,LH*'
		Get 10 minutes of VBB and LP data from MHC.BK and SAO.BK.
	$cmdname -f 1998.176,11:43 -s 2M BKS.BK.HHZ,HHN,HHE,HLZ,HLN,HLE
		Get 2 minutes of VSP and LG data from BKS.BK.
	$cmdname -f 1998.176,11:43 -s 1H '*.BK.LHZ,LHN,LHE'
		Get 1 hour of LP data from all stations.
";
    exit(0);
}

# Depricated syntax:
# 
# [station_list channels_1 [... channels_n]] 
# where:
#	station_list			(depricated syntax)
#		A comma-separated list of station names.  
#		Wildcard characters (*?) may be used within quotes.
#		ALL or '*' indicates all known stations.
#	channels_1 ... channels_n	(deprecated syntax)
#		One or more comma-separated list of SEED channel names.
#		Wildcard characters (*?) may be used within quotes.
#		ALL or '*' indicates all known channels.

{

####################################################################
# User Configurable Section.
#
# 1.  Station, network, and channel lists.
#
# Create list of standard stations and channels.
# Wildcard expansions for stations and channels will be matched
# against this list.  
#
# This should prevent test stations and SOH channels from being
# matched by wildcard characters.  Any unlisted channel can still be
# retrieved by explicitly referencing the desired station and channel.

    $libdir   = "/usr/local/lib";
    $chanfile = "qdata.channels";
    $netsfile = "qdata.nets";
    $DFLT_NET = "BK";
    $SDELIM = ",";
    $CDELIM = ":";
    $NODATA = "";

####################################################################
# Channels that have no instrument response, and that should be
# excluded with the -R option.
####################################################################
    %reject_sncls = ( );

########################################################################
# Pathnames of utilities associated with data extraction.
########################################################################
    $caldate = "/usr/local/bin/caldate";
    $qmerge = "/usr/local/bin/qmerge";

####################################################################
# End User Configurable Section.
####################################################################

    select(STDERR); $| = 1;
    select(STDOUT); $| = 1;

########################################################################
#	Parse command line.
########################################################################
    # Setup info for default data type selection.
    $cmdname = ($0 =~ m|^.*/([^/]+)$|) ? $1 : $0;
    $opt_O = 1;
    $stdfmt = "%Y.%j";

    &getopts ('pnaCBEPOTIAhd:f:t:s:D:o:R:l:');
    &print_syntax($0) if ($opt_h);
    $libdir = $opt_l if ($opt_l ne "");
    $netspath = "$libdir/$netsfile";
    $chanpath = "$libdir/$chanfile";
    # Read initialization files.

    require "$netspath";
    @wc_list = &read_chan_file ($chanpath);
    @dflt_stations = &get_stations ($DFLT_NET);

    # Parse channel list from command line.
    my(@sncl_list) = sort(&parse_channel_spec (@ARGV));

    $current = $opt_C;
    $date_fmt = ($opt_D eq "") ? "j" : $opt_D;
    die "Invalid date format: $opt_D" if (! ($date_fmt =~ /^[jm]$/i));
    goto most_recent if ($opt_P);

    die "Missing -f option or argument." if ($opt_f eq "");
    die "Missing -t or -s option." if ($opt_t eq "" && $opt_s eq "");
    die "Specify only -t or -s option." if ($opt_t ne "" && $opt_s ne "");
    die "Error: specified both -T and -I." if ($opt_T && $opt_I);
    $opt_T = "-T" if (! $opt_I);
    my($all_data_sources) = $opt_a;

    # Create rejection associative array.
    if ($opt_R ne "") {
	my(@reject) = split(/,/,$opt_R);
        my($reject);
	foreach $reject (@reject) {
		my(@sncl) = split(/\./,$reject);
		push(@sncl,"") while (@sncl < 4);
		$reject = join(".",@sncl);
		$reject_sncls{$reject} = 1;
	}
    }

########################################################################
#	Compute normalized representation of start time, end time,
#	and timetag for naming the output files.
########################################################################
    my($timetag);
    my($qt_f) = Qtime::set_time($opt_f);
    die "Error: invalid from time: $opt_f\n" unless (defined($qt_f));
    $timetag = ($newstyle) ? 
	&format_date ($qt_f, "%Y.%j.%H"."%M"."%S") :	# Prevent SCCS subsitutions
	&format_date ($qt_f, "%Y.%j.%H"."%M");		# Prevent SCCS subsitutions
    my($start_datetime) = &caldate("$opt_f");
    my($end_datetime) = ($opt_s) ? 
	&caldate ("$opt_f", "+$opt_s") : 
	&caldate ("$opt_t");
    my($end_time) = &caldate ("-f", $stdfmt, "$end_datetime", "-1F");

########################################################################
#	For each network, find the most recent (chronlogical) date for:
#		archive directories 
#		stage directories.
########################################################################
most_recent: ;
    my($sncl);
    my(%most_recent_date, %archive_incr, %first_prelim, %prelim_start, %prelim_end);
    my(%archive_start, %archive_end);
    foreach $sncl (@sncl_list) {
	my($net) = $sncl =~ /^\w+\.(\w+)\./;
	my($span) = &get_attr("span",$net);
	print (STDERR "Error: unknown attributes for: $net\n"), next if ($span eq "");
	my($start_date) = &start_date($start_datetime,$span);
	if ((! $current) || $opt_P ) {
	    $most_recent_date{$net} = &get_attr("mrecent",$net);
	    $most_recent_date{$net} = &set_most_recent ($net) if ($most_recent_date{$net} == 0);
	    $archive_incr{$net} = $span;
	    if ($opt_P) {
		$first_prelim{$net} = &add_time ($most_recent_date{$net}, $archive_incr{$net}, 1);
		$first_prelim{$net} = &caldate ("-D", $date_fmt, $first_prelim{$net});
		print ($first_prelim{$net}, "\n");
		exit(0);
	    }
	}
	else {
	    $most_recent_date{$net} = &caldate ($start_date, "-1d");
	}

########################################################################
#	Determine whether we want data later than the most recent date
#	on the archive or stage directories.
#	If so, we will have to try to dynamically extract it with qdata.
#	We don't want to extract data if it is not necessary.
########################################################################

	print ("Start date = $start_date, End date = $end_time\n") if ($opt_d);
	if ($most_recent_date{$net} >= $end_time && (! $current) && (! $all_data_sources)) {
	    # No data extraction is required!
	    $prelim_start{$net} = "";
	    $prelim_end{$net} = "";
	    $archive_start{$net} = $start_date;
	    $archive_end{$net} = &add_time ($end_time, $span, 0);
	    print "No data extraction required for network $net.\n" if ($opt_d);
	}
	else {
	    # We need to extract data (drat!).  Compute extraction interval.
	    # Determine whether it ALL needs to be extracted, or whether some
	    # of it is already in the archive or in stage directory.
	    $prelim_end{$net} = $end_time;
	    if ($start_date > $most_recent_date{$net} && (! $all_data_sources)) {
		# All data must come from prelim.
		$prelim_start{$net} = $start_date;
		$archive_start{$net} = "";
		$archive_end{$net} = "";
		print "All data must be extracted for network $net.\n" if ($opt_d);
	    }
	    else {
		# Some data from prelim and extracted data.
		$prelim_start{$net} = &add_time ($most_recent_date{$net}, $span, 1);
		$prelim_end{$net} = $end_time;
		$archive_start{$net} = $start_date;
		$archive_end{$net} = &add_time ($end_time, $span, 0);
		$archive_end{$net} = $most_recent_date{$net} if ($most_recent_date{$net} < $archive_end{$net});
		printf ("Partial data must be extracted for network $net: from %s thru %s\n", 
			$prelim_start{$net}, $prelim_end{$net}) if ($opt_d);
	    }		
	}

########################################################################
#	Construct a list of directories that may contain data files
#	for our timespan.
########################################################################
	&set_data_dirs($net, $archive_start{$net}, $archive_end{$net}, 
		       $prelim_start{$net}, $prelim_end{$net}, $current);

########################################################################
#	Construct command line, 1 for each station channel.
#	Pipe to csh, since they contain wildcard characters.
########################################################################
	if ($opt_p) { open (OUT, ">-"); }
	else { open (OUT,"| /bin/csh -f -s") || die ("Unable to start shell."); }

	my($nreq) = 0;
	my($station,$channel,$loc);
	($station,$net,$channel,$loc) = split(/\./,$sncl);
	my(@prelim_dirs) = &get_attr("prelim",$net);
	my(@dialup_dirs) = &get_attr("dialup",$net);
	my(@dialup_list) = &get_attr("dialdates",$net);
	my(@datelist)    = &get_attr("datelist",$net);

	# Skip rejected channels if requested to do so.
	next if ($opt_R ne "" && &match_reject_sncls($sncl)); 
	my($output) = 
	    ($opt_n) ? "-n" : 
		($opt_o eq "-") ? "" :
		    ($opt_o ne "") ? "-o " . $opt_o :
			"-o " . &output_filename($station,$net,$channel,$loc,$timetag);
	$output .= " -a" if ($opt_o ne "-" && $opt_o ne "" && $nreq > 0);
	$debug = $opt_d | (($opt_A) ? 2048 : 0);
	$debug = ($debug) ? "-d $debug" : "";
	my($cmd) = sprintf ("$qmerge -D $date_fmt $opt_T %s %s -f %s -t %s ", $debug, 
			$output, $start_datetime, $end_datetime);
	my($nfiles) = 0;
	my($paths) = "";
	my($dir);

	# Search prelim directories if needed.
	if ($prelim_start{$net}) {
	    my($cdir);
	    foreach $dir (@prelim_dirs) {
		$cdir = "$dir/$station.$net/$channel.$loc.D";
		if (-d $cdir) {
		    $paths .= "$cdir/* "; 
		    $nfiles++;
		}
	    }
	    # If no prelim directory for this channel, skip dialup.
	    if ($nfiles > 0) {
		my($ts, $dir);
		foreach $dir (@dialup_dirs) {
		    foreach $ts (@dialup_list) {
			$paths .= "$dir/$station.$net.$channel*.D.${ts}* ";
		    }
		}
	    }
	}

	# Search extracted hierarchy for data.
	if ($archive_start{$net}) {
	    my($date, $path);
	    foreach $date (@datelist) {
		$path = &find_file ($station, $net, $channel, $loc, $date);
		if ($path ne "") {
		    $paths .= "$path ";
		    $nfiles++;
		}
	    }
	}

	$cmd .= $paths;
	printf (OUT "%s\n", $cmd), ++$nreq if ($nfiles > 0);
    }
    close (OUT);
}

########################################################################
#	Utility routines.
########################################################################

########################################################################
# get_attr - return specified attribute for the specified network.
########################################################################
sub get_attr {
    my($attr,$net) = @_;
    # Configuration info.
    return (@{$netinfo{$net}{archive}}) if ($attr eq "archive");
    return (@{$netinfo{$net}{dialup}}) if ($attr eq "dialup");
    return (@{$netinfo{$net}{prelim}}) if ($attr eq "prelim");
    return (@{$netinfo{$net}{alt}}) if ($attr eq "alt");
    return ($netinfo{$net}{span}) if ($attr eq "span");
    return ($netinfo{$net}{locinname}) if ($attr eq "locinname");
    # Runtime info.
    return ($netinfo{$net}{mrecent}) if ($attr eq "mrecent");
    return (@{$netinfo{$net}{datelist}}) if ($attr eq "datelist");
    return (@{$netinfo{$net}{datadirs}}) if ($attr eq "datadirs");
    return (@{$netinfo{$net}{dialdates}}) if ($attr eq "dialdates");
    warn ("Error: Unknown attribute: $attr for net: $net");
    return ();
}

########################################################################
# set_most_recent - set and return the most recent date for the network data,
#	based on the archive timespan.
########################################################################
sub set_most_recent {
    my($net) = @_;
    print ("$cmdname - Error: Network $net unknown.\n"), return() if (! defined($netinfo{$net}));
    my($mrecent) = &get_attr("mrecent",$net);
    return ($mrecent) if ($mrecent != "");
    my($dir,$recent);
    # Scan archive directories.
    my($span) = &get_attr("span",$net);
    my(@archive) = &get_attr("archive",$net);
    foreach $dir (@archive) {
	{
	    $recent = &most_recent_year ($dir), last if ($span eq "year");
	    $recent = &most_recent_year_month ($dir), last if ($span eq "month");
	    $recent = &most_recent_year_doy ($dir), last if ($span eq "doy");
	    warn ("Unknown archive timespan: $span"), return ();
	}
	my($recentdate) = &date_from_dir($recent,$span);
	print "Most recent dir in dir $dir = $recent ($recentdate)\n" if ($opt_d);
	$mrecent = $recentdate if ($recentdate gt $mrecent);
    }
    $netinfo{$net}{mrecent} = $mrecent;
    return ($mrecent);
}

########################################################################
# most_recent_year_doy - search for the most recent year and
#	doy directory given the directory hierarchy name.
########################################################################
sub most_recent_year_doy {
    my($parent_dir) = @_;
    opendir (PARENT, $parent_dir) || warn ("$cmdname - Error: Unable to open dir $parent_dir");
    my(@years) = sort (grep (/^\d+$/,readdir(PARENT)));
    my($doy) = "";
    while ($#years >= 0) {
	my($year) = pop ( @years );
	$doy = &most_recent_doy_dir("$parent_dir/$year");
	return ($doy) if ($doy ne "");
    }
    return($doy);
}

########################################################################
# most_recent_doy_dir - search for the most recent 
#	doy directory given the directory hierarchy name.
########################################################################
sub most_recent_doy_dir {
    my($parent_dir) = @_;
    opendir (PARENT, $parent_dir) || warn ("Unable to open dir $parent_dir");
    my(@years) = sort (grep (/^\d+\.\d+$/,readdir(PARENT)));
    my($doy) = pop (@years);
    return($doy);
}

########################################################################
# most_recent_year_month - search for the most recent year and
#	month directory given the directory hierarchy name.
########################################################################
sub most_recent_year_month {
    my($parent_dir) = @_;
    opendir (PARENT, $parent_dir) || warn ("$cmdname - Error: Unable to open dir $parent_dir");
    my(@years) = sort (grep (/^\d+$/,readdir(PARENT)));
    my($month) = "";
    while ($#years >= 0) {
	my($year) = pop ( @years );
	$month = &most_recent_month_dir("$parent_dir/$year");
	return ($month) if ($month ne "");
    }
    return($month);
}

########################################################################
# most_recent_month_dir - search for the most recent 
#	month directory given the directory hierarchy name.
########################################################################
sub most_recent_month_dir {
    my($parent_dir) = @_;
    opendir (PARENT, $parent_dir) || warn ("Unable to open dir $parent_dir");
    my(@years) = sort (grep (/^\d+\.\d+$/,readdir(PARENT)));
    my($month) = pop (@years);
    return($month);
}

########################################################################
# most_recent_year_dir - search for the most recent year 
#	directory given the directory hierarchy name.
########################################################################
sub most_recent_year {
    my($parent_dir) = @_;
    opendir (PARENT, $parent_dir) || warn ("$cmdname - Error: Unable to open dir $parent_dir");
    my(@years) = sort (grep (/^\d+$/,readdir(PARENT)));
    my($year) = pop ( @years );
    return($year);
}

########################################################################
# set_data_dirs - set a list of data directories that could contain
#	data files for our requested timespan.
#	These are either yearly directories that may have daily
#	or monthly subdirectories (for data archive by day or month)
#	or may be directories not ending in a year that may have files.
########################################################################
sub set_data_dirs {
    my($net,$archive_start,$archive_end,$prelim_start,$prelim_end,$current) = @_;
    my(@data_dirs, @dialup_list, $start_day,$end_day);
    my($span) = &get_attr("span",$net);

    ####################################################################
    # Derive list of data file directories to search based on timespan.
    # Dialup needs 1 day before and after for added protection.
    # $dialup_list{$net}:	list of possible year.doy files to find.
    # $data_list{$net}:		list of possible data directories to find.
    ####################################################################
    if ($prelim_start) {
	$start_day = &add_time (&date_from_dir($prelim_start,"doy"),"doy", -1);
	$end_day = &date_from_dir($prelim_end,"doy");
	while ($start_day <= $end_day) {
	    push (@dialup_list, $start_day);
	    $start_day = &add_time ($start_day, $span, 1);
	}
	push (@dialup_list, $start_day);
    }

    my(%datelist,%yearlist);
    if ($archive_start) {
	my($start_unit) = $archive_start;
	my($end_unit) = $archive_end;
	my($year);
	while ($start_unit <= $end_unit) {
	    $datelist{$start_unit} = 1;
	    ($year) = $start_unit =~ /^(\d{4,4})/;	    
	    $yearlist{$year} = 1;
	    $start_unit = &add_time ($start_unit, $span, 1);
	}
    }

    # Derive eligible years from the data_list.
    
    ####################################################################
    # For each extracted unit of time, find data directories that may
    # contain the data.  Search according to the hierarchy defined above.
    # Result will be the list @data_dirs.
    ####################################################################

    # Create list of directories that may contain data files.
    if (! $current) {
	print "Searching extracted data hierarchy ...\n" if ($opt_d);
	my($dir,$ydir,$ent,$year);

	my(@dirs) = &get_attr("archive",$net);
DIR1:	foreach $dir (@dirs) {
	    opendir(DIR, $dir) || warn ("Error: Unable to open dir $dir");
	    my(@years) = grep (/^\d{4,4}$/,readdir(DIR));
	    closedir (DIR);
YEAR1:	    foreach $year (@years) {
		if ($yearlist{$year}) {
		    # This is a year that we want.
		    if ($span eq "year") {
			push (@data_dirs, "$dir/$year");
			next YEAR1;
		    }
		    if ($span eq "month") {
			my($dir2) = "$dir/$year";
			opendir (DIR2, "$dir2") || warn ("Error: unable to open dir $dir2");
			my(@months) = grep(/^$year\.(\d){2,2}$/,readdir(DIR2));
			closedir(DIR2);
			my($month);
			foreach $month (sort @months) {
			    my($datemonth) = &date_from_dir($month,$span);
			    push (@data_dirs, "$dir2/$month") if ($datelist{$datemonth});
			}
			next YEAR1;
		    }
		    if ($span eq "doy") {
			my($dir2) = "$dir/$year";
			opendir (DIR2, "$dir2") || warn ("Error: unable to open dir $dir2");
			my(@doys) = grep(/^$year\.(\d){3,3}$/,readdir(DIR2));
			closedir(DIR2);
			my($doy);
			foreach $doy (sort @doys) {
			    push (@data_dirs, "$dir2/$doy") if ($datelist{$doy});
			}
			next YEAR1;
		    }
		    warn "Error - unknown span: $span";
		}
	    }
	}
    }
    printf ("Data dirs are: %s\n", join (" ",@data_dirs)) if ($opt_d);
    printf ("Dialup dates are: %s\n", join (" ",@dialup_list)) if ($opt_d);
    $netinfo{$net}{datadirs} = [ @data_dirs ];    
    $netinfo{$net}{dialdates} = [ @dialup_list ];
    $netinfo{$net}{datelist} = [ sort(keys(%datelist)) ];
}

########################################################################
# date_from_dir	- convert directory name to date based on span value.
#	doy:	yyyy.doy (or yy.doy).
#	month:	yyyy.mm
#	year:	yyyy
########################################################################
sub date_from_dir {
    my($dir,$span) = @_;
    my($entry) = $dir;
    my($year, $rest, $date, $status);
    $entry = $1 if ($dir =~ m|^.*/([^/]+)$|);
    return ($entry) if ($entry eq "");
    ($year,$rest) = ($entry =~ /^(\d+)(?:\.(\d+))$/);
    $year = ($year >= 50) ? $year+1900 : $year+2000 if ($year < 100);
    my($qt);
    {
	$qt = Qtime::set_time("$year.$rest"), last if ($span eq "doy");
	$qt = Qtime::set_time("$year/$rest/01"), last if ($span eq "month");
	$qt = Qtime::set_time("$year.001"), last if ($span eq "year");
	die "Invalid span: $span";
    }
    die "Invalid dir for span $span\: $dir" unless (defined($qt));
    $date = &format_date($qt, &date_span_format($span));
    return($date);
}

########################################################################
# dir_from_date	- convert date to directory name.
# 	Returns dir name with 4 digit year.
#	Assumption is that date is in full normal form
#		yyyy.ddd[,hh:mm:ss.ffff]
#		yyyy/mm/dd[,hh:mm:ss.ffff]
########################################################################
sub dir_from_date {
    my($date,$span) = @_;
    return ("") if ($date eq "");
    my($dir);
    my($qt) = Qtime::set_time ($date);    
    my($span_fmt) = &dir_span_fmt($span);
    {
	last if ($span eq "doy");
	last if ($span eq "month");
	last if ($span eq "year");
	die "Invalid span: %s";
    }
    $date = &format_date($qt, &dir_span_format($span));
    return ($dir);
}

########################################################################
# find_file - search extracted hierarchy for a file.
#	This routine converts a date to a filename.  
#	NOTE: IT MUST BE CHANGED IF FILENAME OR PATHNAME CONVENTIONS CHANGE.
########################################################################
sub find_file {
    my($station, $net, $channel, $loc, $date) = @_;
    my($dir, $subdir, $scomp, $filename);
    
    my($locinname) = &get_attr("locinname",$net);
    my($loccomp) = ".$loc" if ($locinname);
    my($span) = &get_attr("span",$net);
    my($qt) = Qtime::set_time($date);
    die "Error: Invalid date: $date" unless (defined($qt));
    my($doy) = Qtime::doy($qt);    
    my($month) = Qtime::month($qt);
    my($year) = Qtime::year($qt);
    $year = ($year >= 50) ? $year+1900 : $year+2000 if ($year < 100);
    # The path is is dependent on the network archive timespan.
    # The filename is dependent on the network filename attributes.
    {
	$filename = sprintf ("$station.$net.$channel$loccomp.D.%04d.%03d", $year, $doy),
	    $subdir = sprintf ("%04d.%03d", $year, $doy),
	    $scomp="", 
	    last if ($span eq "doy");
	$filename = sprintf ("$station.$net.$channel$loccomp.D.%04d.%02d", $year, $month),
	    $subdir = sprintf ("%04d.%02d", $year, $month),
	    $scomp="", 
	    last if ($span eq "month");
	$filename = sprintf ("$station.$net.$channel$loccomp.D.%04d", $year),
	    $subdir = "",
	    $scomp = "", 
	    last if ($span eq "year");
	warn ("Error: Unknown timespan: $span");
    }
    
    # Look for file in the extracted hierarchies.
    # The file will be in daily directory within one of the extracted
    # directories.  Use the first file that matches.

    print ("Looking for file $scomp$filename in data directories.\n") if ($opt_d);
    my(@dirs) = &get_attr("datadirs",$net);
    foreach $dir (@dirs) {
	my($dyear,$ydoy,$ymon,$path);
	# If the directory ends in a year, only check for matching year.
	($dyear) = $dir =~ m|/(\d{4,4})$|;
	($ydoy)  = $dir =~ m|/(\d{4,4}\.\d{3,3})$|;
	($ymon)  = $dir =~ m|/(\d{4,4}\.\d{2,2})$|;
	next if (! ($dyear eq $subdir || $ydoy eq $subdir || $ymon eq $subdir));
	$path = "$dir/$scomp$filename";
	print "looking for $path ...\n" if ($opt_d);
	if (-f $path) {
	    print "$filename found ...\n" if ($opt_d);
	    return ($path);
	}
    }
    print "$filename not found ...\n" if ($opt_d);
    return ("");
}

########################################################################
#   output_file:
#	Generate output filename.
########################################################################
sub output_filename {
    my($station,$net,$channel,$loc,$timetag) = @_;
    my($filename);
    my($locinname) = &get_attr("locinname",$net);
    my($loccomp) = ".$loc"; # if ($locinname);
    if ($opt_o) {
	$filename = $opt_o;
    }
    elsif ($newstyle) {
	$filename = "$station.$net.$channel.$loc.D.$timetag";
#	$filename = "$net.$station.$channel.$loc.D.$timetag";
    }
    else {
	$filename = "$station.$channel.D.$timetag";
    }
    return ($filename);
}

########################################################################
# is_leap	- determine whether year is a leap year.
########################################################################
sub is_leap {
    my($year) = @_;		# arguments - year
    return ( $year%400==0 || ($year%4==0 && $year%100!=0) );
}

########################################################################
# add_time	- add N unit to date, and compute new normalized date.
########################################################################

sub add_time {
    my($date,$units,$incr) = @_;	# arguments - date, units, increment
    my($year,$day,$limit);
    my($qt) = Qtime::set_time($date);
    die "Invalid date: $date" unless (defined($qt));
    {
	Qtime::set_doy($qt,Qtime::doy($qt)+$incr), last if ($units eq "doy");
	Qtime::set_month($qt,Qtime::month($qt)+$incr), last if ($units eq "month");
	Qtime::set_year($qt,Qtime::year($qt)+$incr), last if ($units eq "year");
    }    
    my($result) = &format_date ($qt, $stdfmt);
    return ($result);
}

########################################################################
# most_recent_year_day_dir - search for the most recent year and
#	day directory given the directory hierarchy name.
########################################################################
sub most_recent_year_day_dir {
    my($parent_dir) = @_;
    my($day) = "";
    opendir (PARENT, $parent_dir) || die "Unable to open dir $parent_dir";
    my(@years) = sort (grep (/^\d+$/,readdir(PARENT)));
    while ($#years >= 0) {
	my($year) = pop ( @years );
	$day = &most_recent_day_dir("$parent_dir/$year");
	return ($day) if ($day ne "");
    }
    return($day);
}

########################################################################
# most_recent_day_dir - search for the most recent 
#	day directory given the directory hierarchy name.
########################################################################
sub most_recent_day_dir {
    my($parent_dir) = @_;
    opendir (PARENT, $parent_dir) || die "Unable to open dir $parent_dir";
    my(@years) = sort (grep (/^\d+\.\d+$/,readdir(PARENT)));
    my($day) = pop (@years);
    return($day);
}

########################################################################
# expand - generate unique list from a comma-delimited
# 	list of potentially wildcarded strings.
########################################################################
sub expand {
    my($item_string, @match_list) = @_;
    my($item, $str, $pat, @items, %items);
    # First break list into discrete elements.
    $item_string =~ tr/a-z/A-Z/;
    @items = split(/,/,$item_string);
    
    # Loop through each item string, and perform wildcard matching.
    # Wildcard expansion is done against the @match_list array by 
    # substituting shell wildcard characters (?*) with corresponding 
    # regular expression strings, and matching regular expressions.
    foreach $item (@items) {
	$item = "*" if ($item eq "ALL");
	if ($item =~ /.*[\*\?].*/) {
	    ($pat = $item) =~ s/\?/./g;	# Replace ?
	    $pat =~ s/\*/.*/g;	# Replace *
	    foreach $str (@match_list) {
		$items{$str} = $str if ($str =~ /^$pat$/i);
	    }
	}
	else { $items{$item} = $item; }
    }
    @items = sort keys(%items);
    return (@items);
}

########################################################################
#   read_chan_file:
#	Read channel file, return channel list for wildcard use.
#	@wc_list
########################################################################
sub read_chan_file {
    my($datafile) = @_;
    my(@wc_list);
    open (DATAFILE, "$datafile") || 
	warn ("WARNING: wildcard channel list file $datafile not found");
    chomp (@wc_list = grep(!/^(#.*|\s*)$/,<DATAFILE>));
    close (DATAFILE);
    grep (s/\s+//g, @wc_list);
    return (@wc_list);
}

########################################################################
#   get_stations:
#	Return a list of all stations for the specified network.
#	Global variables:	@wc_list
########################################################################
sub get_stations {
    my($net) = @_;
    my($chan,@dflt,%dflt);
    @dflt = grep(/^\w+\.$net\./,@wc_list);
    grep (s/\..*//, @dflt);
    foreach $chan (@dflt) {
	$dflt{$chan} = 1;
    }
    @dflt = sort(keys(%dflt));
    return (@dflt);
}

########################################################################
#   parse_channel_spec:
#	Parse old or new style channel specifier.
#	Return new style channel specifier list with wildcards expanded.
########################################################################
sub parse_channel_spec {
    my(@argv) = @_;
    my($token,$dummy);
    foreach $token (@argv) {
	$dummy = ($token =~ m/\./) ? ++$newstyle : ++$oldstyle;
    }
    die "Error: Mixing old and new style data channel specifiers.\n"
	if ($oldstyle && $newstyle);
    die "Error: Missing data channel specifiers.\n"
	if (! ($oldstyle || $newstyle));
    @argv = &old_to_new(@argv) if ($oldstyle);
    @argv = &gen_new_list(@argv);
    # Return only a sorted unique list.
    my(%argv,$argv);
    foreach $argv (@argv) {
	$argv{$argv} = 1;
    }
    @argv = sort(keys(%argv));
    return (@argv);
}

########################################################################
#   old_to_new:
#	Convert old argument list to new argument list.
#	Return new argument list.
########################################################################
sub old_to_new {
    my(@argv) = @_;
    my(@newlist,@stations,@channels,@channel_list,$channel_list);
    my($station,$channel); 
    @stations = &expand (shift(@argv), @dflt_stations);
    $channel_list = join($CDELIM,@argv);
    @channel_list = split(/$CDELIM/,$channel_list);
    foreach $station (@stations) {
	foreach $channel (@channel_list) {
	    push (@newlist, "$station.$DFLT_NET.$channel.-,\*");
	}
    }
    return (@newlist);
}

########################################################################
#   gen_new_list:
#	Generate new channel specifier list from wildcard specifier list.
#	Return new channel specifier list.
########################################################################
sub gen_new_list {
    my(@argv) = @_;
    my(@list,$argv);
    foreach $argv (@argv) {
	push(@list, &expand_newchan ($argv));
    }
    return (@list);
}

########################################################################
#   expand_newchan:
#	Expand new channel specifier against known channel list.
#	Return matching channel specifier(s).
#	Allow station.net.channel to match station.net.channel.
#	Global:	@wc_list
########################################################################
sub expand_newchan {
    my($arg) = @_;
    $arg .= "." if ($arg =~ /^([^.]+\.[^.]+\.[^.]+)$/);
    return ($arg) if (! $arg =~ m/[\,\?\*]/);
    # Pick apart components, and create outer product from 
    # comma-delimted tokens.
    my($station,$network,$channel,$location);
    ($station,$network,$channel,$location) = $arg =~ /^([^.]+)\.([^.]+)\.([^.]+)\.([^.]*)$/; 
    $location = "-" if ($location eq "");
    my(@stations) = split(/,/,$station);
    my(@networks) = split(/,/,$network);
    my(@channels) = split(/,/,$channel);
    my(@locations) = split(/,/,$location);
    my(@argv,$chan,$pattern);
    foreach $station (@stations) {
	foreach $network (@networks) {
	    foreach $channel (@channels) {
		foreach $location (@locations) {
		    $location =~ s/-//g;
		    $chan = "$station.$network.$channel.$location";
		    print ("Expanded channel: $chan matches:") if ($opt_d);
		    # Accept channel if it doesn't contain wildcard characters.
		    if (! ($chan =~ /[\*\?]/)) {
			push (@argv, $chan);
			print (" $chan", "\n") if ($opt_d);
			next;
		    }
		    # Perform wildcard matching against known channels.
		    ($pattern = $chan) =~ s/\./\\./g;
		    $pattern =~ s/\?/./g;
		    $pattern =~ s/\*/.*/g;
		    foreach $chan (@wc_list) {
			if ($chan =~ m/^($pattern)$/i) {
			    push (@argv, $1);
			    print (" $1") if ($opt_d);
			}
		    }
		    print ("\n") if ($opt_d);
		}
	    }
	}
    }
    return (@argv);
}

#########################################################################
#   caldate: perl implementation of caldate program.			#
#########################################################################
sub caldate {
    my(@args) = @_;
    my($jformat) = "%Y.%j,%H:%M:%S.%F";
    my($mformat) = "%Y/%m/%d,%H:%M:%S.%F";
    my($format) = $jformat;
    while (@args > 0 && $args[0] =~ /^-.+/) {
	my($arg) = shift(@args);
	if ($arg eq "-D") {
	    die "Error: Missing argument to caldate -D" if (@args <= 0);
	    $arg = shift (@args);
	    {
	     	$format = $mformat, last if ($arg eq "m");
	     	$format = $jformat, last if ($arg eq "j");
		die "Error: Invalid option: -D $arg to caldate";
	    }
	}
	elsif ($arg eq "-f") {
	    die "Error: Missing argument to caldate -f" if (@args <= 0);
	    $format = shift (@args);
	}
	elsif ($arg eq "-s") {
	    die "Error: caldate -s not implemented yet";
	}
	elsif ($arg eq "-i" || $arg eq "-d") {
	    die "Error: caldate -i and caldate -d not implemented yet";
	}
    }

    # Get the date.
    die "Error: Missing date to caldate" if (@args <= 0);
    my($date) = shift(@args);
    # Get local or UTC time if requested.
    if ($date =~ /^[lu]$/) {
	my(@t) = ($date eq "u") ? gmtime() : localtime();
	$date = sprintf ("%04d,%03d,%02d:%02d:%02d", 
		$t[5]+1900,$t[7],$t[2],$t[1],$t[0]);
    }
    my($qt) = Qtime::set_time($date);
    die "Error: Bad time in caldate: $date" unless (defined($qt));

    # Perform any date modifications.
    while (@args > 0) {
	my($arg) = shift(@args);
   	my($op,$num,$type);
	($op,$num,$type) = $arg =~ /^([+-]?)(\d*)([yYmdHMSFU]?)$/;
	die "Error: Invalid caldate date modifier: $arg" 
	    if ($op eq "" && $num eq "" && $type eq "");
	$num = 1 if ($num eq "");
	$num *= -1 if ($op eq "-");
	$type = "S" if ($type eq "");
	{
	    Qtime::set_year($qt,Qtime::year($qt)+$num), last if ($type =~ /y/i);
	    Qtime::set_doy($qt,Qtime::doy($qt)+$num), last if ($type eq 'd');
	    Qtime::set_month($qt,Qtime::month($qt)+$num), last if ($type eq 'm');
	    Qtime::set_hour($qt,Qtime::hour($qt)+$num), last if ($type eq 'H');
	    Qtime::set_minute($qt,Qtime::minute($qt)+$num), last if ($type eq 'M');
	    Qtime::set_isecond($qt,Qtime::isecond($qt)+$num), last if ($type eq 'S');
	    Qtime::set_usec($qt,Qtime::usec($qt)+($num*100)), last if ($type eq 'F');
	    Qtime::set_usec($qt,Qtime::usec($qt)+$num), last if ($type eq 'U');
	    die "Error: Invalid caldate date modifier: $arg";
	}
    }
    return(&format_date($qt,$format));
}

#########################################################################
#   format_date: Return a formatted Qtime according to caldate specs.	#
#########################################################################
sub format_date {
    my($qt,$format) = @_;
    my(@fmt) = split(//,$format);
    my($str,$c);
    while (@fmt > 0) {
	$c = shift(@fmt);
	$str .= $c, next if ($c ne '%');
	$c = shift(@fmt);
	die "Error: bad format in format_date: $format" if ($c eq "");
	{
	    $str .= sprintf ("%03d", Qtime::doy($qt)), last if ($c eq 'j');
	    $str .= sprintf ("%04d", Qtime::year($qt)), last if ($c eq 'Y');
	    $str .= sprintf ("%02d", Qtime::year($qt) % 100), last if ($c eq 'y');
	    $str .= sprintf ("%02d", Qtime::month($qt)), last if ($c eq 'm');
	    $str .= sprintf ("%02d", Qtime::day($qt)), last if ($c eq 'd');
	    $str .= sprintf ("%02d", Qtime::hour($qt)), last if ($c eq 'H');
	    $str .= sprintf ("%02d", Qtime::minute($qt)), last if ($c eq 'M');
	    $str .= sprintf ("%02d", Qtime::second($qt)), last if ($c eq 'S');
	    $str .= sprintf ("%04d", Qtime::usec($qt)/100), last if ($c eq 'F');
	    $str .= sprintf ("%06d", Qtime::usec($qt)), last if ($c eq 'U');
	    $str .= $c, last if ($c eq '%');
	}
    }
    return ($str);
}

sub dir_span_format {
    my($span) = @_;
    my($fmt);
    {
	$fmt = "%Y.%j", last if ($span eq "doy");
	$fmt = "%Y.%m", last if ($span eq "month");
	$fmt = "%Y", last if ($span eq "year");
	die "Error: Bad span_format: $span";
    }
    return ($fmt);
}

sub date_span_format {
    my($span) = @_;
    my($fmt);
    {
	$fmt = "%Y.%j", last if ($span eq "doy");
	$fmt = "%Y.%j", last if ($span eq "month");
	$fmt = "%Y.001", last if ($span eq "year");
	die "Error: Bad span_format: $span";
    }
    return ($fmt);
}

sub start_date {
    my($date,$span) = @_;
    my($qt) = Qtime::set_time($date);
    die "Error: Invalid date: $date" unless (defined($qt));
    my($dirdate) = &format_date ($qt,&dir_span_format($span));
    $date = &date_from_dir($dirdate, $span);
    return ($date);
}

sub match_reject_sncls {
    my($sncl) = @_;
    my(@sncl) = ($sncl);
    my($pattern, $reject, $chan);
    foreach $reject (%reject_sncls) {
	# Match reject patterns again this sncl.
	($pattern = $reject) =~ s/\./\\./g;
	$pattern =~ s/\?/./g;
	$pattern =~ s/\*/.*/g;
	foreach $chan (@sncl) {
	    if ($chan =~ m/^($pattern)$/i) {
		return (1);
	    }
	}
    }
    return (0);
}
